from math import floor

import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate as I


class Driver:
    """
    A Driver is typically used to control a certain parameter over time
    for example - the thickness of a line etc.

    In certain respects it's very similar to a Motion object but I think
    it's different enough to warrant its own class

    Attrbiutes
    ----------

    name : string, optional
        The name associated with a particular driver
        Default: ''
    keyframes: [(int/float, float)], optional
        These represent the desired values and times at which they should
        be attained. Time can either be given by the frame number (int) or
        the 'real time' measured in seconds (float). If you wish to specify
        a whole unit of time e.g. 5s it **must** be written as 5.0, otherwise
        it will be interpreted as the frame number 5.
        Default: [(0.0, 0), (1.0, 1)]
    cycle : bool, optional
        If True, any requests for indicies outside the range of the data,
        will be mapped back into it - effectively looping the data.
        Default: False
    FPS : int, optional
        Represents the number of frames and hence data points in each
        second. Default: 25
    """

    def __init__(self, name=None, keyframes=None, cycle=False, FPS=25):
        self._FPS = FPS
        self._cycle = cycle
        self._name = '' if name is None else name

        if keyframes is None:
            self._keys = [(0.0, 0), (1.0, 1)]
        else:
            self._keys = keyframes

        self._calculate()

    def __repr__(self):
        s = "Driver: %s\n\nKeyframes:\n" % self._name

        for key in self._sorted_keys():
            s += "\t%i\t%.2f\t%.2fs\n" %\
                             (key[0], key[1], key[0] / self._FPS)

        s += "\nFPS: %i" % (self._FPS)

        return s

    def _get_single(self, idx):
        """
        This provides the logic for getting a single value
        at a point in time. There are two possible behaviors
        depending on the value of the cycle property.

        If the index is within the length of the _data array
        then this behaves as expected and returns the value at
        that index.

        If however the index is out of bounds of the array then:
            - If cycle is not set and the index is larger than the
              length of the array, then the final value is returned
            - If cycle is not set and the index is less than 0,
              then the first value is returned
            - If cycles is set then the value of index is shifted
              until it lies within the array and that value is retured.
              This effectively extends the array indefinitely. Useful
              for cyclical motion, e.g. a pendulum.
        """
        length = len(self._data)

        # Possibly a bad idea, but we will override the default ability
        # to access from the end of an array by using negative indices
        if idx < 0:
            if not self._cycle:
                return self._data[0]
            else:
                return self._get_single(idx % length)

        # If the index is not negative, just try using it normally
        try:
            return self._data[idx]
        except IndexError:

            # Deal with the consequences of being outside the range
            if not self._cycle:
                return self._data[-1]
            else:
                return self._get_single(idx % length)

    def _get_slice(self, key):
        """
        This implements the ability to get a slice of the
        interpolation data.

        Again depending on if cycle is set we get different
        behavior:
            - If the slice is within the bounds of the array, then
              this behaves as expected.
            - If the slice extends outside of either end of the array
              then:
              + If cycle is set, relevant copies of the values are
                repeated
              + If cycle is not set, the final/first value is repeated
                forever.
        """
        # First unpack the slice to know where we want to slice to and
        # from
        return self._data[key]

    def __getitem__(self, key):
        if isinstance(key, (int,)):
            return self._get_single(key)

        if isinstance(key, (float,)):
            idx = floor(key * self._FPS)
            return self._get_single(idx)

        if isinstance(key, (slice,)):
            return self._get_slice(key)

    def _sorted_keys(self):

        def convert_key(key):
            time, *rest = key

            if isinstance(time, (float,)):
                time = floor(time * self._FPS)

            return tuple([time, *rest])

        return sorted([convert_key(key) for key in self._keys],
                      key=lambda i: i[0])

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):

        if not isinstance(value, (str,)):
            raise TypeError('Name must be a string!')

        self._name = value

    @property
    def frames(self):
        return self._data

    @property
    def keyframes(self):
        return self._keys

    @property
    def cycle(self):
        return self._cycle

    @cycle.setter
    def cycle(self, value):

        if not isinstance(value, (bool,)):
            raise TypeError('Cycle property must be a boolean!')

        self._cycle = value

    @property
    def FPS(self):
        return self._FPS

    @FPS.setter
    def FPS(self, value):

        if not isinstance(value, (int,)):
            raise TypeError('FPS must be an integer!')

        if value <= 0:
            raise ValueError('FPS must be a positive integer!')

        self._FPS = value
        self._calculate()

    def _calculate(self):
        """
        This is the method responsible for doing all the
        interpolation.
        """

        keys = self._sorted_keys()

        # Loop through and interpolate
        self._data = np.array([])

        for idx, key in enumerate(keys[:-1]):
            nextkey = keys[idx + 1]
            newvalues = np.linspace(key[1], nextkey[1],
                                    (nextkey[0] - key[0]))

            if len(newvalues) == 1:
                self._data = np.append(self._data, newvalues)
            else:
                self._data = np.append(self._data, newvalues)[:-1]

        self._data = np.append(self._data, np.array([keys[-1][1]]))

    def show(self):

        fig, ax = plt.subplots(1)
        ax.plot(self._data)
        ax.set_xlabel('Time (in frames)')
        ax.set_ylabel('Value')


def driver(cycle=False, FPS=25):
    """
    This is a decorator which streamlines the process of creating a driver.
    Simply define a function with the times and values you want and the
    decorator does the rest using Python kung-fu!
    """
    def wrapper(f):
        keys = f()

        if keys is None:
            constants = f.__code__.co_consts
            keys = list(filter(lambda i: isinstance(i, (tuple,)), constants))

        return Driver(name=f.__name__, keyframes=keys, cycle=cycle, FPS=FPS)

    return wrapper


class Motion:
    """
    **Note:** This class is not intended to be used directly.
    In order to implement a particular type of motion create
    a new class and inherit from this one.

    The way the motion simulation works is that you specify
    the initial position of the object (x0,y0) and the initial
    velocity of the object (vx0, vy0).

    Then given functions which describe the acceleration in both
    the x and y directions the simulation will integrate them
    up until the current time to get the velocities and positions
    at the current instance.

    Therefore since the intial conditions are all handled by this
    class, in order to implement a new type of motion, inherit from
    this class and override the `_acceleration_x` and
    `_acceleration_y` methods.

    Parameters
    ----------
    These are all the parameters that will be common to any
    motion simualtion:

    x0 : float
        The x coordinate of the starting position of the object.
        Changing this value will trigger the simulation.
        Default: 0
    y0 : float
        The y coordinate of the starting position of the object.
        Changing this value will trigger the simulation.
        Default: 1

    vx0 : float
        The initial x component of velocity of the object.
        Changing this value will trigger the simulation.
        Default: 1
    vy0 : float
        The intial y component of the velocity of the object.
        Changing this value will trigger the simulation.
        Default: 0

    t : int
        This is the time (in seconds) that the simulaion starts
        at. Changing this value will trigger the simulation.
        Default: 0s
    T : int
        This is the time (in seconds) that the simulation will
        finish at. Changing this value will trigger the simulation
        Default: 1s
    FPS: int
        This sets the resolution of the motion, it should match
        the value for the animation. Changing this value will trigger
        the simulation. Default: 25

    Attributes
    ----------
    These are all the attributes that are common to any motion
    simulation

    x : [float]
        This is a list of the x coordinates of the object at each time
        instance
    y : [float]
        This is a list of the y coordinates of the object at each time
        instance
    p : [(float, float)]
        A list of the x-y coordinate of the object at each time instance

    vx : [float]
        A list of the velocities in the x direction at each time
        instance
    vy : [float]
        A list of the velocities in the y direction at each time
        instance
    v : [(float, float)]
        A list of the velocity of the object at each time instance

    ax : [float]
        List of the acceleration in the x direction at each time
        instance
    ay : [float]
        List of the acceleration in the y direction at each time
        instance
    a : [(float, float)]
        List of the acceleration of the object at each time instance
    """

    def __init__(self, x0=0, y0=1, vx0=1, vy0=0, t=0, T=1, FPS=25):
        self._x0 = x0
        self._y0 = y0
        self._vx0 = vx0
        self._vy0 = vy0

        self._t = t
        self._T = T
        self._FPS = FPS

    def __repr__(self):
        s = f"x0:\t{self._x0}\n"
        s += f"y0:\t{self._y0}\n"
        s += f"vx0:\t{self._vx0}\n"
        s += f"vy0:\t{self._vy0}\n"
        s += "\n"
        s += f"t:\t{self._t}\n"
        s += f"T:\t{self._T}\n"
        s += f"FPS:\t{self._FPS}\n"

        return s

    def _velocity_x(self, t):
        return I.quad(lambda s: self._acceleration_x(s), self._t, t)[0]\
               + self._vx0

    def _velocity_y(self, t):
        return I.quad(lambda s: self._acceleration_y(s), self._t, t)[0]\
               + self._vy0

    def _position_x(self, t):
        return I.quad(lambda s: self._velocity_x(s), self._t, t)[0]\
               + self._x0

    def _position_y(self, t):
        return I.quad(lambda s: + self._velocity_y(s), self._t, t)[0]\
               + self._y0

    @property
    def x0(self):
        return self._x0

    @x0.setter
    def x0(self, value):
        self._x0 = value
        self._simulate()

    @property
    def y0(self):
        return self._y0

    @y0.setter
    def y0(self, value):
        self._y0 = value
        self._simulate()

    @property
    def vx0(self):
        return self._vx0

    @vx0.setter
    def vx0(self, value):
        self._vx0 = value
        self._simulate()

    @property
    def vy0(self):
        return self._vy0

    @vy0.setter
    def vy0(self, value):
        self._vy0 = value
        self._simulate()

    @property
    def t(self):
        return self._t

    @t.setter
    def t(self, value):

        if value < 0:
            raise ValueError('Start time cannot be negative!')

        if value >= self._T:
            raise ValueError('Start time cannot be later than end time!')

        self._t = value
        self._simulate()

    @property
    def T(self):
        return self._T

    @T.setter
    def T(self, value):

        if value <= 0:
            raise ValueError('End time must be positive!')

        if value <= self._t:
            raise ValueError('End time cannot be before start time!')

        self._T = value
        self._simulate()

    @property
    def FPS(self):
        return self._FPS

    @FPS.setter
    def FPS(self, value):

        if not isinstance(value, (int,)):
            raise ValueError('FPS must be an integer!')

        if value <= 0:
            raise ValueError('FPS must be positive!')

        self._FPS = value
        self._simulate()

    @property
    def x(self):
        return [x for (x, _, _, _, _, _) in self._data]

    @property
    def y(self):
        return [y for (_, y, _, _, _, _) in self._data]

    @property
    def p(self):
        return list(zip(self.x, self.y))

    @property
    def vx(self):
        return [vx for (_, _, vx, _, _, _) in self._data]

    @property
    def vy(self):
        return [vy for (_, _, _, vy, _, _) in self._data]

    @property
    def v(self):
        return list(zip(self.vx, self.vy))

    @property
    def ax(self):
        return [ax for (_, _, _, _, ax, _) in self._data]

    @property
    def ay(self):
        return [ay for (_, _, _, _, _, ay) in self._data]

    @property
    def a(self):
        return list(zip(self.ax, self.ay))

    def _acceleration_x(self, t):
        return 0

    def _acceleration_y(self, t):
        return 0

    def _simulate(self):

        self._data = []
        frames = np.linspace(self._t, self._T, (self._T - self._t) * self._FPS)

        for t in frames:

            # Compute the _accelerations
            ax = self._acceleration_x(t)
            ay = self._acceleration_y(t)

            # Compute the velocities
            vx = self._velocity_x(t)
            vy = self._velocity_y(t)

            # Compute the positions
            x = self._position_x(t)
            y = self._position_y(t)

            self._data.append((x, y, vx, vy, ax, ay))

    def show(self, vel=False, acc=False):

        if vel or acc:
            fig, (pos, velacc) = plt.subplots(1, 2, figsize=(12, 4))
        else:
            fig, pos = plt.subplots(1)

        pos.plot(self.x, self.y, label='position')
        pos.set_xlabel(r'$x$')
        pos.set_ylabel(r'$y$')
        pos.legend(loc=0)

        if vel:
            velacc.plot(self.vx, label=r'$v_x$')
            velacc.plot(self.vy, label=r'$v_y$')

        if acc:
            velacc.plot(self.ax, label=r'$a_x$')
            velacc.plot(self.ay, label=r'$a_y$')

        if vel or acc:
            velacc.set_xlabel(r'$t$')
            velacc.legend(loc=0)


class Projectile(Motion):
    """
    This implements a very basic notion of projectile motion.
    We assume the following:
      - The object experiences no drag
      - The object has acceleration given by the constant g
    """

    def __init__(self, g=-9.8, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._g = g
        self._simulate()

    def __repr__(self):
        s = "Projectile Motion\n"
        s += "-----------------\n"
        s += f"g:\t{self._g}\n"
        s += '\n'
        s += super().__repr__()

        return s

    def _acceleration_y(self, t):
        return self.g

    @property
    def g(self):
        return self._g

    @g.setter
    def g(self, value):
        self._g = value
        self._simulate()



from .coords import cartesian, extend_periodically, translate, reflect,\
                    polar, mk_domain
from .color import hex_to_rgb
from .image import Image, LayeredImage, TiledImage
from .motion import Projectile, driver
from .objects import TileSet
from .prims import between, circle, ellipse, rectangle, square, thicken,\
                   Puppet
from .time import animate



from math import floor

from .motion import Driver, driver


def make_driver(name, value):
    """
    Given a name and a default value, construct and return
    a Driver with a single keyframe
    """

    @driver()
    def default():
        return [(0.0, value)]

    default.name = name
    return default


def make_property(name):
    """
    Given the name of the property, construct and return a property
    object which points to '_name' and performs appropriate validation
    on the value given to the setter
    """

    def getter(self):
        return self.__getattribute__('_' + name)

    def setter(self, value):

        if not isinstance(value, (Driver,)):
            raise TypeError("%s must be an instance of Driver!" % name)

        # To ensure consistency across the Puppet, set the driver's FPS
        # to match that of the puppet
        value.FPS = self._FPS

        self.__setattr__('_' + name, value)

    return property(fget=getter, fset=setter)


class PuppetMaster(type):
    """
    The PuppetMaster builds puppets.

    This is my first MetaClass, so there's bound to be some problems with
    this eventually but here goes...

    The whole point of creating a MetaClass is to allow for the following
    usage of a class inheriting from Puppet

    class Pacman(Puppet):

        mouth = 0.5
        direction = 0.0    # Angle in radians

        def draw(self, frame, time):

            @cartesian()
            @translate(r=self.direction[frame])
            @polar
            def pacman(x, y, r, t):

                return r <= 0.8 and not\
                      between(-0.6*self.mouth[frame], t, 0.6*self.mouth[frame])

            @pacman.colormap
            def color():
                return (255, 255, 0)

            return pacman

    Note how in the draw() method we used the mouth and direction attributes
    as if they supported indexing? This is precisely what I want the MetaClass
    for. Behind the scenes when Pacman is created using the PuppetMaster class
    which takes the default values and creates a Driver object and hides it
    behind a property object.

    This allows you to write your 'Puppets' exactly as you would if you were
    animating them but you can give them a default pose, and simply plug in
    the animation later hopefully making them much easier to use and re-use.
    """
    def __new__(cls, name, bases, attrs):
        """
        Here is where we do the dirty work, this is called on any
        class definition which uses this as its MetaClass and let's
        us effectively rewrite it before any objects are actually
        produced.

        The main argment of interest to us is the attrs argument
        it's a dictionary of attributes-to-be names and their associated
        values. It contains everything from the module the class is defined
        in, to any __init__ methods and the attributes we are actaully
        interested in.
        """

        # Extract all the attributes we want to be interpreted as
        # drivers. I expect this might cause issue eventually - selecting
        # items which are not meant to be drivers...
        defaults = [(key, val) for key, val in attrs.items()
                    if not key.startswith('_') and
                           isinstance(val, (int, float))]

        # Start rebuilding the attrs argument, starting with everything
        # we don't need to mess with
        new_attrs = [(key, val) for key, val in attrs.items()
                     if (key, val) not in defaults]

        # For certain tasks, it is useful to maintain a list of drivers
        drivers = []

        # Now for each default value
        for key, val in defaults:

            # Record the driver name
            drivers.append('_' + key)

            # Turn the default value into a driver and put it in
            # self._key
            default = make_driver(key, val)
            new_attrs.append(('_' + key, default))

            # Now make a property object to access this 'hidden' driver
            # as well perfrom some sanity checking on any user provided
            # replacements. Store it in self.key
            prop = make_property(key)
            new_attrs.append((key, prop))

        # Add the list of drivers to the list of attributes
        new_attrs.append(('_drivers', drivers))

        # Now that we've rewritten everything we wanted to, tell Python
        # to go ahead and construct the object.

        # I'm not 100% clear on the specifics of this bit, but in order
        # to support inheriting from Puppet, and still have this MetaClass
        # process the definition we need to call super().__new__ rather
        # than type.__new__()
        return super().__new__(cls, name, bases, dict(new_attrs))


class Puppet(metaclass=PuppetMaster):
    """
    A Puppet allows you to easily define time dependent Drawables
    which are one of the building blocks of many animations.

    **Note:** You should NOT use this class directly, but instead
    create your own class and inherit from this one.
    """

    def __init__(self, FPS=25):
        self._FPS = FPS

    def __getitem__(self, key):

        if isinstance(key, (int,)):
            frame = key
            time = key / self._FPS
        elif isinstance(key, (float,)):
            time = key
            frame = floor(key * self._FPS)
        else:
            raise TypeError('Index must be a single timestamp or '
                            'frame number')

        return self.draw(frame, time)

    @property
    def FPS(self):
        return self._FPS

    @FPS.setter
    def FPS(self, value):

        if not isinstance(value, (int,)):
            raise TypeError('FPS must be an integer!')

        if value <= 0:
            raise ValueError('FPS must be positive!')

        self._FPS = value

        # Ensure that the change in FPS is propogated down
        # to the appropriate drivers
        for d in self._drivers:
            self.__getattribute__(d).FPS = value

    def draw(self, frame, time):
        """
        Users, should provide their own implementation of this method
        """
        raise NotImplementedError('Puppets must have a draw method!!')


def ellipse(x0, y0, a, b, r, pt=0.2, fill=False):
    """
    An ellipse can be defined in the following manner:

        1/a(x - x0)^2 + 1/b(y - y0)^2 = r^2

    This function returns a function in (x, y) and will
    return true if the given point is in the ellipse

    Arguments:
    ----------

    x0 : float
        Represents the x coordinate of the ellipse's center
    y0 : float
        Represents the y coordinate of the ellipse's center
    a : float
        This controls how elongated the ellipse is in the x
        direction
    b : float
        This controls how elongated the ellipse is in the y
        direction
    r : float
        This represents the radius of the ellipse
    pt : float, optional
        Represents the thickness of the lines of the ellipse.
        Default: 0.2
    fill : bool, optional
        Fill the ellipse rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt
    """

    lhs = lambda x, y: 1/a*(x - x0)**2 + 1/b*(y - y0)**2

    if fill:

        def ell(x, y):

            if lhs(x, y) <= r**2:
                return True

            return False

        return ell

    else:

        def ipse(x, y):

            val = lhs(x, y)

            if val <= (r + pt)**2 and val >= (r - pt)**2:
                return True

            return False
        return ipse


def circle(x0, y0, r, *args, **kwargs):
    """
    A circle can be defined in the following manner:

        (x - x0)^2 + (y - y0)^2 = r^2

    This function returns a function in (x, y) and
    will return true if the point is in the circle

    Arugments:
    ----------

    x0 : float
        This is the x coordinate of the circle's center
    y0 : float
        This is the y coordinate of the circle's center
    r : float
        This represents the radius of the ellipse
    pt : float, optional
        Represents the thickness of the lines of the circle.
        Default: 0.2
    fill : bool, optional
        Fill the circle rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt
    """

    return ellipse(x0, y0, 1, 1, r, *args, **kwargs)


def thicken(f, pt):
    """
    Mathematical paths are all well and good but in order
    to see them they have to have some amount of thickness.

    Given some function which describes a path and a thickness
    this function construct and returns a function which determines
    if some given y is in the neighbourhood of the given function
    at x

    Arguments:
    ----------

    f: callable
        The function you want 'thickened'
    pt: float
        Determines the size of the neighbourhood to test
        in
    """

    def test_func(x, y):

        if y >= f(x) - pt and y <= f(x) + pt:
            return True
        else:
            return False

    return test_func


def between(lower, value, upper):
    return lower <= value and value <= upper


def rectangle(x0, y0, width, height, pt=0.2, fill=False):
    """
    Returns a function in (x, y) which returns true
    when the given point in in/on the rectangle defined by the
    arguments.

    Arguments:
    ----------

    x0 : float
        Represents the x-coordinate of the rectangle's center
    y0 : float
        Represents the y-coordinate of the rectangle's center
    width : float
        Represents the width of the rectangle
    height : float
        Represents the height of the rectangle
    pt : float, optional
        Represents the thickness of the lines of the rectangle.
        Default: 0.2
    fill : bool, optional
        Fill the rectangle rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt
    """
    left = x0 - (width / 2)
    right = x0 + (width / 2)
    top = y0 + (height / 2)
    bottom = y0 - (height / 2)

    def rect(x, y):

        if x >= left and x <= right and\
           y >= bottom and y <= top:
            return True

        return False

    if fill:
        return rect

    def small(x, y):

        if x >= left + pt and x <= right - pt and\
           y >= bottom + pt and y <= top - pt:
            return True

        return False

    def test(x, y):

        if rect(x, y) and not small(x, y):
            return True

        return False

    return test


def square(x0, y0, size, *args, **kwargs):
    """
    Returns a function in (x, y) which returns true
    when the given point in in/on the square defined by the
    arguments.

    Arguments:
    ----------

    x0 : float
        Represents the x-coordinate of the rectangle's center
    y0 : float
        Represents the y-coordinate of the rectangle's center
    size : float
        Represents the size of the rectangle
    pt : float, optional
        Represents the thickness of the lines of the rectangle.
        Default: 0.2
    fill : bool, optional
        Fill the rectangle rather than outline it
        Default: False
        **Note:** If fill is true, this function will ignore the value
        of pt

    """
    return rectangle(x0, y0, size, size, *args, **kwargs)
